"""Custom types for the project."""

import os
import pathlib
from typing import TypeVar

import Quartz

# Acccept any type that can be converted to a path
FilePath = TypeVar("FilePath", str, pathlib.Path, os.PathLike)

# Create a custom type for CGMutableImageMetadataRef
# This is used to indicate which functions require a mutable copy of the metadata
# The Quartz package doesn't provide a CGMutableImageMetadataRef type
CGMutableImageMetadataRef = Quartz.CGImageMetadataRef



"""Classes for CGMetadata"""

import pathlib
from typing import IO, Any, Literal

from .cgmetadata import (
    load_image_metadata_ref,
    load_image_properties,
    load_video_metadata,
    load_video_xmp,
)
from .constants import (
    EXIF,
    GPS,
    IPTC,
    TIFF,
    UDTA,
    XMP,
    XMP_PACKET_FOOTER,
    XMP_PACKET_HEADER,
)
from .metadata import (
    metadata_dictionary_from_image_metadata_ref,
    metadata_ref_create_mutable_copy,
    metadata_ref_set_tag_for_dict,
    metadata_ref_set_tag_with_path,
    metadata_ref_write_to_file,
)
from .types import FilePath
from .utils import (
    CFDictionary_to_dict,
    is_image,
    is_video,
    single_quotes_to_double_quotes,
    strip_xmp_packet,
)
from .xmp import metadata_ref_create_from_xmp, metadata_ref_serialize_xmp


class ImageMetadata:
    """Read and write image metadata properties using native macOS APIs.

    Args:
        filepath: The path to the image file.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If the file is not an image file.
    """

    def __init__(self, filepath: FilePath):
        self.filepath = pathlib.Path(filepath).resolve()
        if not self.filepath.exists():
            raise FileNotFoundError(f"File not found: {self.filepath}")
        if not is_image(self.filepath):
            raise ValueError(f"Not an image file: {self.filepath}")
        self._context_manager = False
        self._load()

    @property
    def properties(self) -> dict[str, Any]:
        """Return the metadata properties dictionary from the image.

        The dictionary keys are named 'IPTC', 'EXIF', etc.
        Some of the values are themselves dictionaries.
        For example, the 'IPTC' value is a dictionary of IPTC metadata.

        Reference: https://developer.apple.com/documentation/imageio/image_properties?language=objc
        for more information.
        """

        # change keys to remove the leading '{' and trailing '}'
        # e.g. '{IPTC}' -> 'IPTC' but only if the key starts with '{'
        # also change Exif -> EXIF, WebP -> WEBP to match the other keys
        properties = CFDictionary_to_dict(self._properties)
        properties = {
            key[1:-1] if key.startswith("{") else key: value
            for key, value in properties.items()
        }
        if "Exif" in properties:
            properties["EXIF"] = properties.pop("Exif")
        if "WebP" in properties:
            properties["WEBP"] = properties.pop("WebP")
        return properties

    @property
    def exif(self) -> dict[str, Any]:
        """Return the EXIF properties dictionary from the image."""
        return self.properties.get(EXIF, {})

    @property
    def iptc(self) -> dict[str, Any]:
        """Return the IPTC properties dictionary from the image."""
        return self.properties.get(IPTC, {})

    @property
    def tiff(self) -> dict[str, Any]:
        """Return the TIFF properties dictionary from the image."""
        return self.properties.get(TIFF, {})

    @property
    def gps(self) -> dict[str, Any]:
        """Return the GPS properties dictionary from the image."""
        return self.properties.get(GPS, {})

    @property
    def xmp(self) -> dict[str, Any]:
        """Return the XMP metadata dictionary for the image.

        The dictionary keys are in form "prefix:name", e.g. "dc:creator".
        """
        return metadata_dictionary_from_image_metadata_ref(self._metadata_ref)

    def xmp_dumps(self, header: bool = True) -> str:
        """Return the serialized XMP metadata for the image.

        Args:
            header: If True, include the XMP packet header in the serialized XMP.

        Returns:
            The serialized XMP metadata for the image as a string.
        """
        xmp = metadata_ref_serialize_xmp(self._metadata_ref).decode("utf-8")
        if header:
            xmp = f"{XMP_PACKET_HEADER}\n{xmp}\n{XMP_PACKET_FOOTER}"
        return xmp

    def xmp_dump(self, fp: IO[str], header: bool = True):
        """Write the serialized XMP metadata for the image to a file.

        Args:
            fp: The file pointer to write the XMP metadata to.
            header: If True, include the XMP packet header in the serialized XMP.
        """
        xmp = metadata_ref_serialize_xmp(self._metadata_ref).decode("utf-8")
        if header:
            xmp = XMP_PACKET_HEADER + xmp + XMP_PACKET_FOOTER
        fp.write(xmp)

    def xmp_loads(self, xmp: str):
        """Load XMP metadata from a string.

        Args:
            xmp: The XMP metadata as a string.

        Note:
            This does not write the metadata to the image file.
            Use write() to write the loaded metadata to the image file.
            The XMP standard allows quoted strings to use either single or double quotes.
            For example, exiftool uses single quotes. However, the native macOS APIs
            (CGImageMetadataCreateFromXMPData) returns nil if the XMP data contains single quotes.
            This does not appear to be documented anywhere in the Apple documentation.
            This function replaces single quotes with double quotes to avoid this issue.
        """
        self._xmp_set_from_str(xmp)

    def xmp_load(self, fp: IO[str]):
        """Load XMP metadata from a file.

        Args:
            fp: The file pointer to read the XMP metadata from.

        Note:
            This does not write the metadata to the image file.
            Use write() to write the loaded metadata to the image file.
            The XMP standard allows quoted strings to use either single or double quotes.
            For example, exiftool uses single quotes. However, the native macOS APIs
            (CGImageMetadataCreateFromXMPData) returns nil if the XMP data contains single quotes.
            This does not appear to be documented anywhere in the Apple documentation.
            This function replaces single quotes with double quotes to avoid this issue.
        """
        xmp = fp.read()
        self._xmp_set_from_str(xmp)

    def set(
        self,
        group: Literal["EXIF", "IPTC", "TIFF", "GPS", "XMP"],
        key: str,
        value: Any,
    ):
        """Set a metadata property for the image.

        Args:
            group: The metadata group type to set the property for, for example, "IPTC", "XMP"
            key: The key or key path of the metadata property to set;
                for "XMP" metadata, the key is in form "prefix:name", e.g. "dc:creator", "dc:description"...
                for other metadata, the key is the name of the property, e.g. "LensModel", "Make", "Keywords"...
            value: The value to set the metadata property to.

        Note:
            This does not write the metadata to the image file unless used in conjunction with the context manager.
            Use write() to write the metadata to the image file after setting one or more values.
            Metadata keys may be specified as a literal string, e.g. "LensModel" or using
            one of the constants from the ImageIO module, e.g. kCGImagePropertyExifLensModel,
            which are referenced here: https://developer.apple.com/documentation/imageio/exif_dictionary_keys
            These are available in the pyobjc Quartz module as Quartz.kCGImagePropertyExifLensModel, etc.
            You are responsible for passing the correct type of value for the metadata key,
            for example, str or list[str]. See https://github.com/adobe/xmp-docs/tree/master
            for more information on XMP metadata and expected types.
        """
        if group == XMP:
            self._metadata_ref = metadata_ref_set_tag_with_path(
                self._metadata_ref, key, value
            )
        else:
            self._metadata_ref = metadata_ref_set_tag_for_dict(
                self._metadata_ref, group, key, value
            )

    def write(self):
        """Write the metadata to the image file then reloads the metadata from the image."""
        metadata_ref_write_to_file(self.filepath, self._metadata_ref)
        self.reload()

    def reload(self):
        """Reload the metadata from the image file."""
        self._load()

    def asdict(self) -> dict[str, Any]:
        """Return the metadata as a dictionary."""
        dict_data = self.properties.copy()
        dict_data[XMP] = self.xmp
        return dict_data

    def _load(self):
        try:
            del self._metadata_ref
        except AttributeError:
            pass
        try:
            del self._properties
        except AttributeError:
            pass
        properties = load_image_properties(self.filepath)
        self._properties = properties.mutableCopy()
        del properties

        metadata_ref = load_image_metadata_ref(self.filepath)
        self._metadata_ref = metadata_ref_create_mutable_copy(metadata_ref)
        del metadata_ref

    def _xmp_set_from_str(self, xmp: str):
        """Set the XMP metadata from a string representing serialized XMP."""

        # The Apple API requires that the XMP data use double quotes for quoted strings
        # and that the XMP data not contain the XMP packet headers
        xmp = single_quotes_to_double_quotes(xmp)
        xmp = strip_xmp_packet(xmp)
        xmp = xmp.encode("utf-8")
        self._xmp_set_from_bytes(xmp)

    def _xmp_set_from_bytes(self, xmp: bytes):
        """Set the XMP metadata from a bytes object representing serialized XMP."""
        metadata = metadata_ref_create_from_xmp(xmp)
        self._metadata_ref = metadata_ref_create_mutable_copy(metadata)
        del metadata

    def __enter__(self):
        """Enter the context manager."""
        self._context_manager = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        if self._context_manager:
            self.write()
            self.reload()
        self._context_manager = False

    def __del__(self):
        if self._metadata_ref is not None:
            del self._metadata_ref
        if self._properties is not None:
            del self._properties


class ImageMetaData(ImageMetadata):
    """Alias for ImageMetadata."""

    pass


class VideoMetadata:
    """Read video metadata properties using native macOS APIs.

    Args:
        filepath: The path to the video file.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If the file is not an video file.

    Note: Unlike ImageMetadata, this class does not provide write access to the metadata.
    """

    def __init__(self, filepath: FilePath):
        self.filepath = pathlib.Path(filepath).resolve()
        if not self.filepath.exists():
            raise FileNotFoundError(f"File not found: {self.filepath}")
        if not is_video(self.filepath):
            raise ValueError(f"Not a video file: {self.filepath}")
        self._context_manager = False
        self._load()

    @property
    def properties(self) -> dict[str, Any]:
        """Return the metadata properties dictionary from the image.

        The dictionary keys are named with the namespace, such as 'mdta', 'udta'.
        Some of the values are themselves dictionaries.
        """
        return self._properties

    @property
    def xmp(self) -> dict[str, Any]:
        """Return the XMP metadata dictionary for the image.

        The dictionary keys are in form "prefix:name", e.g. "dc:creator".
        """
        return self._properties.get(XMP, {})

    def xmp_dumps(self, header: bool = True) -> str:
        """Return the serialized XMP metadata for the video.

        Args:
            header: If True, include the XMP packet header in the serialized XMP.

        Returns:
            The serialized XMP metadata for the image as a string.
        """
        xmp = self._xmp
        if not header:
            xmp = strip_xmp_packet(xmp)
        return xmp

    def xmp_dump(self, fp: IO[str], header: bool = True):
        """Write the serialized XMP metadata for the video to a file.

        Args:
            fp: The file pointer to write the XMP metadata to.
            header: If True, include the XMP packet header in the serialized XMP.
        """
        xmp = self.xmp_dumps(header)
        xmp = xmp.encode("utf-8")
        fp.write(xmp)

    def reload(self):
        """Reload the metadata from the image file."""
        self._load()

    def asdict(self) -> dict[str, Any]:
        """Return the metadata as a dictionary."""
        dict_data = self._properties.copy()
        return dict_data

    def _load(self):
        try:
            del self._properties
        except AttributeError:
            pass
        self._properties = load_video_metadata(self.filepath)
        self._xmp = load_video_xmp(self.filepath)


class VideoMetaData(VideoMetadata):
    """Alias for VideoMetadata."""

    pass


# class XMPMetadata:
#     """Read and write image metadata properties using native macOS APIs.

#     Args:
#         filepath: The path to the image file or the XMP sidecar file.
#             If None, the metadata will be initialized with an empty dictionary.
#     """

#     def __init__(self, filepath: FilePath | None = None):
#         self.filepath = filepath
#         self.init_with_image = is_image(filepath) if filepath else False
#         self.init_with_xmp = not self.init_with_image and filepath is not None
#         self._metadata_ref = self._load_metadata_ref()

#     @property
#     def metadata(self):
#         """Get the XMP metadata dictionary for the image.

#         The dictionary keys are in form "prefix:name", e.g. "dc:creator".
#         """
#         return self._metadata_ref

#     def serialize(self, header: bool = True) -> str:
#         """Return the serialized XMP metadata for the image.

#         Args:
#             header: If True, include the XMP packet header in the serialized XMP.

#         Returns: The serialized XMP metadata for the image as a string.
#         """
#         xmp = metadata_ref_serialize_xmp(self._metadata_ref)

#         xmp = xmp.decode("utf-8")
#         if header:
#             xmp = XMP_PACKET_HEADER + xmp

#         return xmp

#     def _load_metadata_ref(self) -> CGMutableImageMetadataRef:
#         """Called by __init__ to load the metadata reference."""
#         if self.init_with_image:
#             metadata = load_image_metadata_ref(self.filepath)
#             metadata_ref = metadata_ref_create_mutable_copy(metadata)
#             del metadata
#         elif self.init_with_xmp:
#             xmp = open(self.filepath, "r").read()
#             xmp = xmp.encode("utf-8")
#             metadata = metadata_ref_create_from_xmp(xmp)
#             metadata_ref = metadata_ref_create_mutable_copy(metadata)
#             del metadata
#         else:
#             metadata_ref = metadata_ref_create_empty_mutable()
#         return metadata_ref

#     def __del__(self):
#         if self._metadata_ref is not None:
#             del self._metadata_ref



"""Utilities for working with CGMetadataRef"""

from __future__ import annotations

from typing import Any, Literal

import objc
import Quartz
from CoreFoundation import (
    CFArrayGetCount,
    CFArrayGetTypeID,
    CFArrayGetValueAtIndex,
    CFDictionaryGetTypeID,
    CFDictionaryRef,
    CFGetTypeID,
    CFStringGetTypeID,
)
from Foundation import (
    NSURL,
    NSArray,
    NSData,
    NSDictionary,
    NSMutableArray,
    NSMutableDictionary,
)
from wurlitzer import pipes

from .types import CGMutableImageMetadataRef, FilePath


class MetadataError(Exception):
    """Error calling Quartz.CGImageMetadata functions."""

    pass


def metadata_ref_create_mutable_copy(
    metadata_ref: Quartz.CGImageMetadataRef | CGMutableImageMetadataRef,
) -> CGMutableImageMetadataRef:
    """Create a CGMutableImageMetadataRef copy from a Quartz.CGImageMetadataRef."""
    with objc.autorelease_pool():
        return Quartz.CGImageMetadataCreateMutableCopy(metadata_ref)


def metadata_ref_create_empty_mutable() -> CGMutableImageMetadataRef:
    """Create an empty CGMutableImageMetadataRef."""
    with objc.autorelease_pool():
        return Quartz.CGImageMetadataCreateMutable()


def metadata_ref_set_tag_with_path(
    metadata_ref: CGMutableImageMetadataRef,
    tag_path: str,
    value: Any,
) -> CGMutableImageMetadataRef:
    """Set a metadata tag to value in a CGMutableImageMetadataRef

    Args:
        metadata_ref: A CGMutableImageMetadataRef
        tag_path: The tag path to set
        value: The value to set

    Returns: CGMutableImageMetadataRef with the tag set to value
    """
    with objc.autorelease_pool():
        if Quartz.CGImageMetadataSetValueWithPath(metadata_ref, None, tag_path, value):
            return metadata_ref
        raise MetadataError(
            f"Could not set tag {tag_path} to {value}; "
            "verify the tag and value are valid and that metadata_ref is a CGMutableImageMetadataRef"
        )


def metadata_ref_set_tag_for_dict(
    metadata_ref: CGMutableImageMetadataRef,
    dictionary: Literal[
        "EXIF",
        "IPTC",
        "XMP",
        "TIFF",
        "GPS",
        "WEBP",
        "HEIC",
        "CIFF",
        "DNG",
        "GIF",
        "JFIF",
        "PNG",
        "TGA",
        "8BIM",
    ],
    tag: str,
    value: Any,
) -> CGMutableImageMetadataRef:
    """Set a metadata tag to value in a CGMutableImageMetadataRef for the given dictionary

    Args:
        metadata_ref: A CGMutableImageMetadataRef
        dictionary: The dictionary to set (e.g. "EXIF", "IPTC", "XMP", "TIFF", "GPS")
        tag: The tag to set
        value: The value to set

    Returns: CGMutableImageMetadataRef with the tag set to value
    """
    dict_names = {
        "EXIF": Quartz.kCGImagePropertyExifDictionary,
        "IPTC": Quartz.kCGImagePropertyIPTCDictionary,
        "GPS": Quartz.kCGImagePropertyGPSDictionary,
        "WEBP": Quartz.kCGImagePropertyWebPDictionary,
        "HEIC": Quartz.kCGImagePropertyHEICSDictionary,
        "CIFF": Quartz.kCGImagePropertyCIFFDictionary,
        "DNG": Quartz.kCGImagePropertyDNGDictionary,
        "GIF": Quartz.kCGImagePropertyGIFDictionary,
        "JFIF": Quartz.kCGImagePropertyJFIFDictionary,
        "PNG": Quartz.kCGImagePropertyPNGDictionary,
        "TGA": Quartz.kCGImagePropertyTGADictionary,
        "TIFF": Quartz.kCGImagePropertyTIFFDictionary,
        "8BIM": Quartz.kCGImageProperty8BIMDictionary,
    }
    dict_name = dict_names.get(dictionary.upper())
    if not dict_name:
        raise MetadataError(f"Invalid dictionary {dictionary}")

    with objc.autorelease_pool():
        if Quartz.CGImageMetadataSetValueMatchingImageProperty(
            metadata_ref, dict_name, tag, value
        ):
            return metadata_ref
        raise MetadataError(
            f"Could not set tag {tag} to {value} for dictionary {dict_name}; "
            "verify the dictionary, tag, and value are valid and that metadata_ref is a CGMutableImageMetadataRef"
        )


def metadata_ref_write_to_file(
    image_path: FilePath,
    metadata_ref: Quartz.CGImageMetadataRef,
) -> None:
    """Write metadata to an image file.

    Args:
        image_path: Path to the image file.
        metadata_ref: A CGImageMetadataRef containing the metadata to write.
    """
    with objc.autorelease_pool():
        image_url = NSURL.fileURLWithPath_(str(image_path))
        image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)
        if not image_source:
            raise MetadataError(f"Could not create image source for {image_path}")
        image_type = Quartz.CGImageSourceGetType(image_source)
        destination = Quartz.CGImageDestinationCreateWithURL(
            image_url, image_type, 1, None
        )
        if not destination:
            raise MetadataError(f"Could not create image destination for {image_path}")
        with pipes() as (_out, _err):
            # On some versions of macOS this causes error to stdout
            # of form: AVEBridge Info: AVEEncoder_CreateInstance: Received CreateInstance (from VT)...
            # even though the operation succeeds
            # Use pipes() to suppress this error
            image_data = Quartz.CGImageSourceCreateImageAtIndex(image_source, 0, None)
            Quartz.CGImageDestinationAddImageAndMetadata(
                destination,
                image_data,
                metadata_ref,
                None,
            )
            Quartz.CGImageDestinationFinalize(destination)

        del destination
        del image_data
        del image_source


def properties_dict_write_to_file(
    image_path: FilePath,
    properties: CFDictionaryRef,
) -> None:
    """Write properties to an image file.

    Args:
        image_path: Path to the image file.
        properties: A CFDictionaryRef containing the properties to write.
    """
    with objc.autorelease_pool():
        image_url = NSURL.fileURLWithPath_(str(image_path))
        image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)
        if not image_source:
            raise MetadataError(f"Could not create image source for {image_path}")
        image_type = Quartz.CGImageSourceGetType(image_source)
        destination = Quartz.CGImageDestinationCreateWithURL(
            image_url, image_type, 1, None
        )
        if not destination:
            raise MetadataError(f"Could not create image destination for {image_path}")
        with pipes() as (_out, _err):
            # On some versions of macOS this causes error to stdout
            # of form: AVEBridge Info: AVEEncoder_CreateInstance: Received CreateInstance (from VT)...
            # even though the operation succeeds
            # Use pipes() to suppress this error
            image_data = Quartz.CGImageSourceCreateImageAtIndex(image_source, 0, None)
            Quartz.CGImageDestinationAddImageFromSource(
                destination, image_source, 0, properties
            )
            Quartz.CGImageDestinationFinalize(destination)

        del destination
        del image_data
        del image_source


def properties_dict_set_tag(
    properties_dict: CFDictionaryRef, sub_dict: str | None, tag: str, value: Any
) -> CFDictionaryRef:
    """Set a tag to value in a CFDictionaryRef as returned by load_image_properties.

    Args:
        properties_dict: A CFDictionaryRef
        sub_dict: The sub dictionary to set or None if the tag is not in a sub dictionary
        tag: The tag to set
        value: The value to set

    Returns: CFDictionaryRef with the tag set to value
    """
    mutable_dict = properties_dict.mutableCopy()
    if sub_dict:
        dict_ref = mutable_dict[sub_dict].mutableCopy()
        if not dict_ref:
            dict_ref = NSMutableDictionary.dictionary()
        dict_ref[tag] = value
        mutable_dict[sub_dict] = dict_ref
        return mutable_dict

    mutable_dict[tag] = value
    return mutable_dict


def NSDictionary_to_dict_recursive(ns_dict: NSDictionary) -> dict[str, Any]:
    """Convert an NSDictionary to a Python dict recursively; handles subset of types needed for image metadata."""
    py_dict = {}
    for key, value in ns_dict.items():
        if isinstance(value, NSDictionary):
            py_dict[key] = NSDictionary_to_dict_recursive(value)
        elif isinstance(value, NSArray):
            py_dict[key] = NSArray_to_list_recursive(value)
        elif isinstance(value, NSData):
            py_dict[key] = value.bytes().tobytes()
        else:
            py_dict[key] = str(value)
    return py_dict


def NSArray_to_list_recursive(ns_array: NSArray) -> list[Any]:
    """Convert an NSArray to a Python list recursively; handles subset of types needed for image metadata."""
    py_list = []
    for value in ns_array:
        if isinstance(value, NSDictionary):
            py_list.append(NSDictionary_to_dict_recursive(value))
        elif isinstance(value, NSArray):
            py_list.append(NSArray_to_list_recursive(value))
        elif isinstance(value, NSData):
            py_list.append(value.bytes().tobytes())
        else:
            py_list.append(str(value))
    return py_list


def metadata_dictionary_from_image_metadata_ref(
    metadata_ref: Quartz.CGImageMetadataRef,
) -> dict[str, Any]:
    with objc.autorelease_pool():
        tags = Quartz.CGImageMetadataCopyTags(metadata_ref)
        if not tags:
            return {}

        metadata_dict = {}
        for i in range(CFArrayGetCount(tags)):
            tag = CFArrayGetValueAtIndex(tags, i)

            prefix = Quartz.CGImageMetadataTagCopyPrefix(tag)
            name = Quartz.CGImageMetadataTagCopyName(tag)
            value = Quartz.CGImageMetadataTagCopyValue(tag)

            key = f"{prefix}:{name}"
            object_value = _recursive_parse_metadata_value(value)
            metadata_dict[key] = object_value

        return metadata_dict.copy()


def _recursive_parse_metadata_value(value: Any) -> Any:
    if CFGetTypeID(value) == CFStringGetTypeID():
        return str(value)
    elif CFGetTypeID(value) == CFDictionaryGetTypeID():
        value_dict = NSMutableDictionary.dictionary()
        original_dict = NSDictionary.dictionaryWithDictionary_(value)
        for key in original_dict.allKeys():
            value_dict[key] = _recursive_parse_metadata_value(original_dict[key])
        return NSDictionary_to_dict_recursive(value_dict)
    elif CFGetTypeID(value) == CFArrayGetTypeID():
        value_array = NSMutableArray.array()
        original_array = NSArray.arrayWithArray_(value)
        for element in original_array:
            value_array.addObject_(_recursive_parse_metadata_value(element))
        return NSArray_to_list_recursive(value_array)
    elif CFGetTypeID(value) == Quartz.CGImageMetadataTagGetTypeID():
        tag_value = Quartz.CGImageMetadataTagCopyValue(value)
        return _recursive_parse_metadata_value(tag_value)
    else:
        return value



"""Utility functions for the project."""

from __future__ import annotations

import re
from typing import Any

import AppKit
import objc
import UniformTypeIdentifiers
from Foundation import NSURL, CFArrayRef, CFDataRef, CFDictionaryRef

from .types import FilePath


def is_image(filepath: FilePath) -> bool:
    """Return True if the file at `filepath` is an image file.

    Args:
        filepath: The path to the file to check.

    Returns: True if the file is an image file, False otherwise.

    Note: This function will work whether or not the file exists as the
    UTI is determined by the file extension.
    """

    return _conforms_to_uti(filepath, "public.image")


def is_video(filepath: FilePath) -> bool:
    """Return True if the file at `filepath` is a video file.

    Args:
        filepath: The path to the file to check.

    Returns: True if the file is an video file, False otherwise.

    Note: This function will work whether or not the file exists as the
    UTI is determined by the file extension.
    """

    return _conforms_to_uti(filepath, "public.movie")


def _conforms_to_uti(filepath: FilePath, uti: str) -> bool:
    """Returns True if filepath confirms to a given UTI otherwise False"""
    # Cycle through all types associated with file and check if any of them are the requested type
    # There are several APIs that would have made this easier but Apple has deprecated them all
    with objc.autorelease_pool():
        url = NSURL.fileURLWithPath_(str(filepath))
        provider = AppKit.NSItemProvider.alloc().initWithContentsOfURL_(url)
        registered_type_identifiers = provider.registeredTypeIdentifiers()
        image_type = UniformTypeIdentifiers.UTType.typeWithIdentifier_(uti)
        for type_id in registered_type_identifiers:
            current_type = UniformTypeIdentifiers.UTType.typeWithIdentifier_(type_id)
            if current_type.conformsToType_(image_type):
                return True
        return False


def single_quotes_to_double_quotes(s: str) -> str:
    """Replace single quotes with double quotes in a string."""

    # this is a bit of hack

    # replace all escaped backslashes with a placeholder
    placeholder = "\u0000\u0001\u0000"
    s = s.replace(r"\\", placeholder)

    # replace all single quotes with double quotes except for escaped single quotes
    s = re.sub(r"(?<!\\)'", '"', s)

    # replace all placeholders with escaped backslashes
    s = s.replace(placeholder, r"\\")

    return s


def strip_xmp_packet(xmp: str) -> str:
    """Strip XMP packet header and footer from string.

    Args:
        xmp: An XMP string.

    Returns: The XMP string with the packet header and footer removed.
    """
    header_pattern = (
        r"<\?xpacket begin=['\"]\ufeff['\"] id=['\"]W5M0MpCehiHzreSzNTczkc9d['\"]\?>"
    )
    footer_pattern = r"<\?xpacket end=['\"]w['\"]\?>"

    xmp = re.sub(header_pattern, "", xmp)
    xmp = re.sub(footer_pattern, "", xmp)

    return xmp.strip()


def cftype_to_pytype(value: Any) -> Any:
    """Convert a Core Foundation type to a python type
    This doesn't cover every type but covers types I've seen in metadata
    """
    if isinstance(value, NSURL):
        value = str(value.path())
    elif isinstance(value, objc.pyobjc_unicode):
        value = str(value)
    elif isinstance(value, CFDataRef):
        value = bytes(value)
    elif isinstance(value, objc._pythonify.OC_PythonLong):
        value = int(value)
    elif isinstance(value, objc._pythonify.OC_PythonFloat):
        value = float(value)
    elif isinstance(value, list):
        value = list(cftype_to_pytype(v) for v in value)
    return value


def CFDictionary_to_dict(cf_dict: CFDictionaryRef) -> dict:
    """Recursively convert a CFDictionary to a dict, converting any objective C types to python equivalent."""
    if cf_dict is None:
        return None
    d = {}
    for key, value in cf_dict.items():
        if isinstance(value, CFDictionaryRef):
            d[key] = CFDictionary_to_dict(value)
        else:
            if isinstance(value, CFArrayRef):
                value = list(cftype_to_pytype(v) for v in value)
            d[key] = cftype_to_pytype(value)
    return d



"""Utilities for working with XMP data"""

from __future__ import annotations

from typing import Any

import objc
import Quartz
from Foundation import NSData, __NSCFData

from .constants import XMP_PACKET_FOOTER, XMP_PACKET_HEADER
from .metadata import MetadataError, metadata_dictionary_from_image_metadata_ref
from .utils import single_quotes_to_double_quotes, strip_xmp_packet


def is_xmp_packet(value: bytes | str | __NSCFData) -> bool:
    """Check if a metadata value is an XMP packet.

    Args:
        value: A metadata value.

    Returns: True if the value is an XMP packet, False otherwise.
    """
    if isinstance(value, (bytes, __NSCFData)):
        value = value.decode("utf-8")
    return value.startswith("<?xpacket begin='\ufeff") or value.startswith(
        '<?xpacket begin="\ufeff'
    )


def metadata_dictionary_from_xmp_packet(
    xmp: str | bytes | __NSCFData,
) -> dict[str, Any]:
    """Extract key-value pairs from an XMP packet.

    Args:
        xmp: str or bytes for XMP packet

    Returns: dict with key/value pairs from the XMP packet.

    Raises:
        ValueError: if XMP packet cannot be decoded.
    """
    with objc.autorelease_pool():
        if isinstance(xmp, (bytes, __NSCFData)):
            xmp = xmp.decode("utf-8")
        xmp = single_quotes_to_double_quotes(xmp)
        xmp = strip_xmp_packet(xmp)
        xmp = xmp.encode("utf-8")
        if mdref := metadata_ref_create_from_xmp(xmp):
            return metadata_dictionary_from_image_metadata_ref(mdref)
        raise ValueError("Failed to create metadata ref from XMP packet")


def metadata_ref_serialize_xmp(metadata_ref: Quartz.CGImageMetadataRef) -> bytes:
    """Create serialized XMP from a Quartz.CGImageMetadataRef."""
    with objc.autorelease_pool():
        data = Quartz.CGImageMetadataCreateXMPData(metadata_ref, None)
        if not data:
            return b""
        xmp = bytes(data)
        del data
        return xmp


def metadata_ref_create_from_xmp(xmp: bytes) -> Quartz.CGImageMetadataRef:
    """Create a Quartz.CGImageMetadataRef from serialized XMP.

    Args:
        xmp: bytes object containing serialized XMP

    Returns: Quartz.CGImageMetadataRef

    Raises:
        MetadataError: If the metadata could not be created from the XMP data.

    Note: Quoted strings in the xmp data must use double quotes, not single quotes (apostrophes).
        The XMP standard allows strings in either format and exiftool uses single quotes but
        CGImageMetadataCreateFromXMPData returns nil if the XMP data contains single quotes.
        This does not appear to be documented anywhere in the Apple documentation.
        Also, the Apple documentation for CGImageMetadataCreateFromXMPData is incorrect;
        it states the XMP may contain the XMP packet header but if provided a valid XMP tree
        with packet header, the function returns nil.

        Reference: https://developer.apple.com/documentation/imageio/1465001-cgimagemetadatacreatefromxmpdata?language=objc
    """
    with objc.autorelease_pool():
        data = NSData.dataWithBytes_length_(xmp, len(xmp))
        metadata_ref = Quartz.CGImageMetadataCreateFromXMPData(data)
        if not metadata_ref:
            raise MetadataError("Could not create metadata from XMP data")
        del data
        return metadata_ref



"""Use native Core Graphics API on macOS to access and change image metadata

This code is an alternative to using a third-party tool like the excellent [exiftool](https://exiftool.org/)
and uses Apple's native ImageIO APIs. It should be able to read metadata from any file format supported
by ImageIO.

Implementation Note: Core Foundation objects created with Create or Copy functions must be released;
this is done with `del Object` in pyobjc.
"""

from __future__ import annotations

from typing import Any

import objc
import Quartz
from AVFoundation import AVURLAsset
from CoreFoundation import CFDictionaryCreate, CFDictionaryRef
from Foundation import NSURL

from .constants import MDTA, UDTA, XMP
from .iso6709 import parse_iso_6709
from .metadata import (
    NSDictionary_to_dict_recursive,
    metadata_dictionary_from_image_metadata_ref,
)
from .types import FilePath
from .xmp import is_xmp_packet, metadata_dictionary_from_xmp_packet


def load_image_properties(
    image_path: FilePath,
) -> CFDictionaryRef:
    """Return the metadata properties dictionary from the image at the given path.

    Args:
        image_path: Path to the image file.

    Returns:
        A CFDictionary dictionary of metadata properties from the image file.
        If the image does not contain metadata, an empty CFDictionary is returned.

    Note:
        The dictionary keys are named '{IPTC}', '{TIFF}', etc.
        Reference: https://developer.apple.com/documentation/imageio/image_properties?language=objc
        for more information.

        This function is useful for retrieving EXIF and IPTC metadata.
    """
    with objc.autorelease_pool():
        image_url = NSURL.fileURLWithPath_(str(image_path))
        image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)

        metadata = Quartz.CGImageSourceCopyPropertiesAtIndex(image_source, 0, None)
        del image_source
        return metadata or CFDictionaryCreate(None, [], [], 0, None, None)


def load_image_properties_dict(
    image_path: FilePath,
) -> dict[str, Any]:
    """Return the metadata properties dictionary from the image at the given path.

    Args:
        image_path: Path to the image file.

    Returns:
        A dictionary of metadata properties from the image file.

    Note:
        The dictionary keys are named '{IPTC}', '{TIFF}', etc.
        Reference: https://developer.apple.com/documentation/imageio/image_properties?language=objc
        for more information.

        This function is useful for retrieving EXIF and IPTC metadata.
    """
    metadata = load_image_properties(image_path)
    return NSDictionary_to_dict_recursive(metadata)


def load_image_metadata_dict(
    image_path: FilePath,
) -> dict[str, Any]:
    """Get the XMP metadata from the image at the given path

    Args:
        image_path: Path to the image file.

    Returns:
        A dictionary of XMP metadata properties from the image file.
        The dictionary keys are in form "prefix:name", e.g. "dc:creator".
    """
    metadata = load_image_metadata_ref(str(image_path))
    return metadata_dictionary_from_image_metadata_ref(metadata) if metadata else {}


def load_image_metadata_ref(
    image_path: FilePath,
) -> Quartz.CGImageMetadataRef:
    """Get the Quartz.CGImageMetadataRef from the image at the given path

    Args:
        image_path: Path to the image file.

    Returns:
        A Quartz.CGImageMetadataRef containing the XMP metadata.
        If the image does not contain metadata, an empty Quartz.CGImageMetadataRef is returned.
    """
    with objc.autorelease_pool():
        image_url = NSURL.fileURLWithPath_(str(image_path))
        image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)

        metadata = Quartz.CGImageSourceCopyMetadataAtIndex(image_source, 0, None)
        del image_source
    return metadata or Quartz.CGImageMetadataCreateMutable()


def load_image_location(
    image_path: FilePath,
) -> tuple[float, float]:
    """Return the GPS latitude/longitude coordinates from the image at the given path.

    Args:
        image_path: Path to the image file.

    Returns:
        A tuple of latitude and longitude.

    Raises:
        ValueError: If the image does not contain GPS data or if the GPS data does not contain latitude and longitude.
    """
    properties = load_image_properties_dict(image_path)
    gps_data = properties.get(Quartz.kQuartz.CGImagePropertyGPSDictionary)
    if not gps_data:
        raise ValueError("This image does not contain GPS data")

    latitude = gps_data.get(Quartz.kQuartz.CGImagePropertyGPSLatitude)
    longitude = gps_data.get(Quartz.kQuartz.CGImagePropertyGPSLongitude)

    if latitude is None or longitude is None:
        raise ValueError("Could not extract latitude and/or longitude from GPS data")

    try:
        latitude = float(latitude)
        longitude = float(longitude)
    except ValueError:
        raise ValueError("Could not extract latitude and/or longitude from GPS data")

    if gps_data.get(Quartz.kQuartz.CGImagePropertyGPSLatitudeRef) == "S":
        latitude *= -1
    if gps_data.get(Quartz.kQuartz.CGImagePropertyGPSLongitudeRef) == "W":
        longitude *= -1

    return latitude, longitude


def load_video_metadata(video_path: FilePath) -> dict[str, Any]:
    """Load metadata from a video file using AVFoundation.

    Args:
        video_path: Path to a video file.

    Returns: A dictionary of metadata key-value pairs.
    """
    with objc.autorelease_pool():
        video_path = str(video_path)
        video_url = NSURL.fileURLWithPath_(video_path)
        asset = AVURLAsset.URLAssetWithURL_options_(video_url, None)

        metadata_formats = asset.availableMetadataFormats()
        metadata_dictionary = {}

        for format in metadata_formats:
            metadata_items = asset.metadataForFormat_(format)

            for item in metadata_items:
                namespace = str(item.keySpace()) if item.keySpace() else ""
                if key := item.commonKey():
                    key = str(key)
                    key = key[0].upper() + key[1:] if len(key) > 1 else key.upper()
                else:
                    key = ""  # I've seen null key
                value = item.value()
                if namespace not in metadata_dictionary:
                    metadata_dictionary[namespace] = {}
                if value is not None:
                    if namespace == UDTA and not key:
                        # user data, possibly an XMP packet
                        if is_xmp_packet(value):
                            metadata_dictionary[XMP] = (
                                metadata_dictionary_from_xmp_packet(value)
                            )
                        else:
                            metadata_dictionary[namespace][key] = str(value)
                    elif namespace == MDTA and key == "Location":
                        try:
                            coordinates = parse_iso_6709(value)
                            metadata_dictionary[namespace][key] = {
                                "Latitude": coordinates[0],
                                "Longitude": coordinates[1],
                                "Height": coordinates[2],
                                "CRS": coordinates[3],
                            }
                        except ValueError:
                            metadata_dictionary[namespace][key] = str(value)
                    else:
                        metadata_dictionary[namespace][key] = str(value)
        return metadata_dictionary


def load_video_xmp(video_path: FilePath) -> str | None:
    """Load XMP metadata packet from a video file using AVFoundation.

    Args:
        video_path: Path to a video file.

    Returns: str containing the XMP metadata packet.
    """
    with objc.autorelease_pool():
        video_path = str(video_path)
        video_url = NSURL.fileURLWithPath_(video_path)
        asset = AVURLAsset.URLAssetWithURL_options_(video_url, None)

        metadata_formats = asset.availableMetadataFormats()

        for format in metadata_formats:
            metadata_items = asset.metadataForFormat_(format)

            for item in metadata_items:
                namespace = str(item.keySpace()) if item.keySpace() else ""
                if key := item.commonKey():
                    key = str(key)
                    key = key[0].upper() + key[1:] if len(key) > 1 else key.upper()
                else:
                    key = ""  # I've seen null key
                value = item.value()
                if value is not None:
                    if namespace == UDTA and not key:
                        # user data, possibly an XMP packet
                        if is_xmp_packet(value):
                            return value.decode("utf-8")
    return None


# def load_image_auxilary_data(image_path: FilePath) -> CFDictionaryRef:
#     """Return the auxiliary data dictionary from the image at the given path."""
#     with objc.autorelease_pool():
#         image_url = NSURL.fileURLWithPath_(str(image_path))
#         image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)

#         aux_data = Quartz.CGImageSourceCopyAuxiliaryDataInfoAtIndex(
#             image_source, 0, Quartz.kCGImageAuxiliaryDataTypeXMP
#         )
#         del image_source
#         return aux_data


# def image_file_write_properties_metadata(
#     image_path: FilePath,
#     properties: CFDictionaryRef,
#     metadata_ref: Quartz.CGImageMetadataRef,
# ) -> None:
#     """Write properties and metadata to an image file."""
#     with objc.autorelease_pool():
#         image_url = NSURL.fileURLWithPath_(str(image_path))
#         image_source = Quartz.CGImageSourceCreateWithURL(image_url, None)
#         if not image_source:
#             raise MetadataError(f"Could not create image source for {image_path}")
#         image_type = Quartz.CGImageSourceGetType(image_source)
#         destination = Quartz.CGImageDestinationCreateWithURL(
#             image_url, image_type, 1, None
#         )
#         if not destination:
#             raise MetadataError(f"Could not create image destination for {image_path}")
#         with pipes() as (_out, _err):
#             # On some versions of macOS this causes error to stdout
#             # of form: AVEBridge Info: AVEEncoder_CreateInstance: Received CreateInstance (from VT)...
#             # even though the operation succeeds
#             # Use pipes() to suppress this error
#             image_data = Quartz.CGImageSourceCreateImageAtIndex(image_source, 0, None)
#             Quartz.CGImageDestinationAddImageAndMetadata(
#                 destination,
#                 image_data,
#                 metadata_ref,
#                 None,
#             )
#             Quartz.CGImageDestinationSetProperties(destination, properties)
#             Quartz.CGImageDestinationFinalize(destination)
#         del image_source
#         del image_data
#         del destination



"""CLI for CGMetadata. """

from __future__ import annotations

import argparse
import csv
import json
import pathlib
import sys
from typing import Any

from rich import print
from rich.console import Console
from rich.table import Table

from ._version import __version__
from .classes import ImageMetadata, VideoMetadata
from .utils import is_image, is_video


def main():
    """Print metadata for image and video files."""
    arg_parser = argparse.ArgumentParser(
        description="Print metadata for image and video files in various formats."
    )
    arg_parser.add_argument(
        "image", metavar="IMAGE_OR_VIDEO", help="path to image or video file"
    )
    arg_parser.add_argument(
        "--version", "-v", action="version", version="%(prog)s " + __version__
    )
    arg_parser.add_argument(
        "--csv",
        "-c",
        action="store_true",
        help="output as comma separated values (CSV); see also --no-header",
    )
    arg_parser.add_argument(
        "--tsv",
        "-t",
        action="store_true",
        help="output as tab separated values (TSV); see also --no-header",
    )
    arg_parser.add_argument(
        "--json", "-j", action="store_true", help="output as JSON; see also --indent"
    )
    arg_parser.add_argument(
        "--xmp",
        "-x",
        action="store_true",
        help="output XMP sidecar for image; see also --no-header",
    )
    arg_parser.add_argument(
        "--indent",
        "-i",
        type=int,
        help="indent level for JSON; default 4, use 0 for no indentation",
        default=4,
    )
    arg_parser.add_argument(
        "--no-header",
        "-H",
        action="store_true",
        help="when used with --csv, --tsv, omit column headers; "
        "when used with --xmp, omit XMP packet header",
    )
    args = arg_parser.parse_args()

    # argument validation
    if not pathlib.Path(args.image).exists():
        print(f"File not found: {args.image}", file=sys.stderr)
        sys.exit(1)
    if not (is_image(args.image) or is_video(args.image)):
        print(f"Not an image or video file: {args.image}", file=sys.stderr)
        sys.exit(1)
    if sum([args.csv, args.tsv, args.json, args.xmp]) > 1:
        print(
            "Only one of --csv, --tsv, --json, --xmp can be specified", file=sys.stderr
        )
        sys.exit(1)

    # load metadata and print in the appropriate format
    if is_image(args.image):
        md = ImageMetadata(args.image)
    else:
        md = VideoMetadata(args.image)
    dict_data = md.asdict()

    if args.json:
        indent = args.indent or None
        json_str = md_dict_to_json(dict_data, indent)
        print(json_str)
        sys.exit(0)

    if args.csv:
        print_csv(dict_data, header=not args.no_header)
        sys.exit(0)

    if args.tsv:
        print_csv(dict_data, delim="\t", header=not args.no_header)
        sys.exit(0)

    if args.xmp:
        print(md.xmp_dumps(header=not args.no_header))
        sys.exit(0)

    # default output
    print_md_dict(dict_data, args.image)
    sys.exit(0)  # always raise SystemExit for testing


def md_dict_to_json(md_dict: dict[str, Any], indent: int = 4) -> str:
    """Convert a dictionary of metadata to a JSON string."""

    def default(o):
        if isinstance(o, bytes):
            return format_bytes(o)
        raise TypeError(
            f"Object of type {o.__class__.__name__} is not JSON serializable"
        )

    return json.dumps(md_dict, default=default, indent=indent)


def format_bytes(b: bytes) -> str:
    """Format bytes as a string for pretty printing."""
    length = len(b)
    first = min(length, 16)
    first_bytes = b[:first].hex()
    value = f"<{length} bytes: {first_bytes}"
    if length > first:
        value += "..."
    value += ">"
    return value


def format_value(value: Any) -> Any:
    if isinstance(value, list):
        return ", ".join(str(v) for v in value)
    if isinstance(value, bytes):
        return format_bytes(value)
    return str(value)


def flatten_metadata_dict(
    d: dict[str, Any], group_key: str = "", parent_key: str = ""
) -> list[tuple[str, str, Any]]:
    """Flatten metadata dictionary into a list of tuples of (group, tag, value)"""
    items = []
    for k, v in d.items():
        if parent_key:
            new_key = f"{parent_key}.{k}"
        else:
            new_key = k
        if isinstance(v, dict):
            if group_key:
                items.extend(flatten_metadata_dict(v, group_key, new_key))
            else:
                items.extend(flatten_metadata_dict(v, new_key))
        else:
            items.append((group_key, new_key, v))
    return items


def print_md_dict(md_dict: dict[str, Any], filename: str):
    """Print metadata dictionary."""

    items = sorted(flatten_metadata_dict(md_dict))
    table = Table(title=f"Metadata for {filename}")
    table.add_column("Group")
    table.add_column("Tag")
    table.add_column("Value")
    for group, tag, value in items:
        group = group or "None"
        table.add_row(group, tag, format_value(value))
        # items_to_print.append((group, tag, format_value(value)))
        # print(f"{group} : {tag} : {format_value(value)}")
    console = Console()
    console.print(table)


def print_csv(md_dict: dict[str, Any], delim: str = ",", header: bool = True):
    """Print metadata dictionary as CSV.

    Args:
        md_dict: metadata dictionary
        delim: delimiter to use
        header: if True, print header row
    """

    items = sorted(flatten_metadata_dict(md_dict))
    items_to_print = []
    for group, tag, value in items:
        group = group or "None"
        items_to_print.append((group, tag, format_value(value)))

    csv_writer = csv.writer(sys.stdout, delimiter=delim)
    if header:
        csv_writer.writerow(["Group", "Tag", "Value"])
    csv_writer.writerows(items_to_print)


if __name__ == "__main__":
    main()



"""Use native Core Graphics / ImageIO API on macOS to access and change image metadata"""

from ._version import __version__
from .classes import ImageMetadata, ImageMetaData, VideoMetadata, VideoMetaData
from .constants import EXIF, GPS, IPTC, MDTA, TIFF, UDTA, XMP
from .metadata import MetadataError
from .xmp import metadata_dictionary_from_xmp_packet

__all__ = [
    "EXIF",
    "GPS",
    "ImageMetadata",
    "ImageMetaData",
    "IPTC",
    "is_xmp_packet",
    "MDTA",
    "metadata_dictionary_from_xmp_packet",
    "MetadataError",
    "TIFF",
    "UDTA",
    "VideoMetadata",
    "VideoMetaData",
    "XMP",
]



"""Constants for CGMetadata."""

# XMP packet header and footer for serializing XMP metadata.
XMP_PACKET_HEADER = '<?xpacket begin="\uFEFF" id="W5M0MpCehiHzreSzNTczkc9d"?>'
XMP_PACKET_FOOTER = '<?xpacket end="w"?>'

# Metadata types
XMP = "XMP"
EXIF = "EXIF"
IPTC = "IPTC"
TIFF = "TIFF"
GPS = "GPS"
WEBP = "WEBP"

# Metadata namespaces
MDTA = "mdta"
UDTA = "udta"



"""Work with ISO 6709 coordinate strings."""

import re


def parse_iso_6709(iso_6709_str: str) -> tuple[float, float, float | None, str | None]:
    """
    Parse an ISO 6709 string and return latitude and longitude as floats.

    The ISO 6709 string should be in the format: '+33.9180-118.3814+039.512/'
    - The first component is latitude with a sign (+ or -).
    - The second component is longitude with a sign (+ or -).
    - The third component is height with a sign (+ or -).
    - The Coordinated Reference System (CRS) is optional (e.g. CRSWGS_84);
        this contravenes the specification but matches actual practice.
    - There may or may not be a trailing slash.

    Args:
        iso_6709_str: The ISO 6709 string to parse.

    Returns:
        A tuple containing the latitude and longitude as floats, the height as a float or None, and the CRS as a string or None.

    Raises:
        ValueError: If the string is not a valid ISO 6709 string.
    """
    pattern = re.compile(
        r"""
        ^  # Start of string
        (?P<latitude>[+-]\d+(?:\.\d+)?)  # Latitude with sign and optional decimal
        (?P<longitude>[+-]\d+(?:\.\d+)?)  # Longitude with sign and optional decimal
        (?:  # Optional height and CRS group
            (?P<height>[+-]?\d+(?:\.\d+)?)  # Height with sign and optional decimal
            # (?:CRS(?P<crs>[^/]+))?  # Optional CRS (string beginning with CRS), only if height is present
            (?:CRS(?P<crs>.*?(?=\/|$)))?  # Optional CRS (string beginning with CRS), only if height is present
        )?
        /?  # Optional trailing "/"
        $  # End of string
    """,
        re.VERBOSE,
    )

    match = pattern.match(iso_6709_str)
    if match:
        latitude = float(match.group("latitude"))
        longitude = float(match.group("longitude"))
        height = float(match.group("height")) if match.group("height") else None
        crs = match.group("crs") if match.group("crs") else None
        if crs and not height:
            raise ValueError(f"CRS cannot be present without height: {iso_6709_str}")
        return latitude, longitude, height, crs
    else:
        raise ValueError(f"Invalid ISO 6709 location string: {iso_6709_str}")
